#!/usr/bin/env python

import os
import sys
import json
import click

from RAMP.packer import package
from RAMP.unpacker import validate_bundle, unpack as unpack_bundle
from RAMP.version import VERSION
import RAMP.module_metadata as module_metadata

def comma_seperated_to_list(ctx, param, value):
    """
    Converts a comma seperated string into a list.
    """
    if value is None:
        return []
    else:
        items = value.split(',')
        return list(set(items))
    
def jsons_str_tuple_to_jsons_tuple(ctx, param, value):
    """
    Converts json str into python map
    """
    if value is None:
        return []
    else:
        return [json.loads(a) for  a in value]

@click.group()
@click.version_option(version=VERSION)
def ramp():
    pass

@ramp.command()
def version():
    print('RAMP packer v={}'.format(str(VERSION)))
    return 0

@ramp.command()
@click.argument('bundle')
def validate(bundle):
    valid, e = validate_bundle(bundle)
    if valid:
        print("package is valid")
        return 0
    else:
        print("package is invalid, reason: %s" % e)
    return 1

@ramp.command()
@click.argument('bundle')
def unpack(bundle):
    metadata, module = unpack_bundle(bundle)
    module_metadata_file_name = os.path.join(os.getcwd(), metadata['module_name'] + '.json')
    module_file_name = os.path.join(os.getcwd(), metadata['module_file'])

    with open(module_metadata_file_name, 'w') as outfile:
        json.dump(metadata, outfile)
        print(module_metadata_file_name)

    with open(module_file_name, 'w') as outfile:
        for line in module.readlines():
            outfile.write(line)
        print(module_file_name)

    return 0

@ramp.command()
@click.argument('module')
@click.option('--manifest', '-m', type=click.File('rb'), help='generate package from manifest')
@click.option('--display-name', '-d', default=None, help='name for display purposes')
@click.option('--module-name', '-n', default=None, help='module name')
@click.option('--author', '-a', default=None, help='module author')
@click.option('--email', '-e', default=None, help='author\'s email')
@click.option('--architecture', '-A', default=None, help='module compiled on i386/x86_64 arch')
@click.option('--description', '-D', default=None, help='short description')
@click.option('--homepage', '-h', default=None, help='module homepage')
@click.option('--license', '-l', default=None, help='license')
@click.option('--cmdargs', '-c', 'command_line_args', default=None, help='module command line arguments')
@click.option('--redis-min-version', '-r', 'min_redis_version', default=None, help='redis minimum version')
@click.option('--redis-pack-min-version', '-R', 'min_redis_pack_version', default=None, help='redis pack minimum version')
@click.option('--config-command', '-cc', default=None, help='command used to configure module args at runtime')
@click.option('--os', '-O', default=None, help='build target OS (Darwin/Linux)')
@click.option('--capabilities', '-C', callback=comma_seperated_to_list, help='comma seperated list of module capabilities')
@click.option('--exclude-commands', '-E', callback=comma_seperated_to_list, help='comma seperated list of exclude commands')
@click.option('--overide-command', multiple=True, callback=jsons_str_tuple_to_jsons_tuple, help='gets a command json representation and overide it on the module json file')

@click.option('--output', '-o', default='module.zip', help='output file name')
@click.option('--print-filename-only', '-P', is_flag=True, default=False, help="Print package path, but don't generate file")
@click.option('--verbose', '-v', is_flag=True, default=False, help='verbose mode: print the resulting metadata')
def pack(module, *args, **kwargs):
    return package(module, **kwargs)

if __name__ == '__main__':
    ramp()



import os
import sys
import json
import zipfile
import hashlib
import semantic_version

from RAMP import module_metadata
from distutils.version import StrictVersion

MAX_MODULE_FILE_SIZE = 1024 * 1024 * 10

class UnpackerPackageError(Exception):
    """
    Represents and error within the unpacking process
    """
    def __init__(self, message, reason=None):
        super(UnpackerPackageError, self).__init__(message)
        self.reason = reason

    def __str__(self):
        if self.reason:
            return "{}, reason: {}".format(super(UnpackerPackageError, self).__str__(), self.reason)
        else:
            return "{}".format(super(UnpackerPackageError, self).__str__())

def _sha256_checksum(module_file):
    """Computes sha256 for given file"""

    sha256 = hashlib.sha256()
    sha256.update(module_file.read())
    return sha256.hexdigest()

def unpack(bundle):
    """
    Unpacks a bundeled module, performs sanity validation
    on bundle.
    :return:
    :rtype: tuple
    both the module metadata and the actual module are returned
    """

    # Extract.
    try:
        with zipfile.ZipFile(bundle) as zf:
            # validate_zip_file throws
            # we want this exception to propagate
            _validate_zip_file(zf)

            try:
                metadata = json.load(zf.open('module.json'))
            except (IOError, ValueError):
                raise UnpackerPackageError("Failed to read module.json")

            module = zf.open(metadata["module_file"])

            # _validate throws incase of an error,
            # we want this exception to propagate
            _validate_metadata(metadata, module)

    except zipfile.BadZipfile:
        raise UnpackerPackageError("Failed to extract bundle")

    return (metadata, module)

def validate_bundle(bundle):
    """
    Checks to see if given bundle is valid.
    Return (True, None) if so, (False, Exception) otherwise.
    """
    try:
        unpack(bundle)
        return (True, None)
    except Exception as e:
        return (False, e)

def _validate_zip_file(zip_file):
    """
    Checks if all entries within the zip file don't
    exceed a certian threshold.
    """
    infolist = zip_file.infolist()
    # We're expecting exactly two files.
    if len(infolist) != 2:
        raise UnpackerPackageError("module zip file did not pass sanity validation", reason="module zip file should contains exactly two file")

    # Check zip content size.
    for zipinfo in infolist:
        # Size of the compressed/uncompressed data.
        if zipinfo.file_size > MAX_MODULE_FILE_SIZE:
            raise UnpackerPackageError("module zip file did not pass sanity validation", reason="module file content is too big")

    return True

def _validate_metadata(metadata, module):
    """
    Checks metadata isn't missing any required fields
    metadata - dictionary
    module - path to module file
    """

    for field in module_metadata.FIELDS:
        if not field in metadata:
            raise UnpackerPackageError("module did not pass sanity validation", reason="Missing mandatory field [{}]".format(field))

    # Empty module name
    if metadata["module_name"] == "":
        raise UnpackerPackageError("module did not pass sanity validation", reason="Empty module name")

    # Empty module file name
    if metadata["module_file"] == "":
        raise UnpackerPackageError("module did not pass sanity validation", reason="Empty module file name")

    # Architecture must 64 bits
    if metadata["architecture"] != "x86_64":
        raise UnpackerPackageError("module did not pass sanity validation", reason="Architecture must 64 bits")

    # Missing version
    if not semantic_version.validate(metadata["semantic_version"]):
        raise UnpackerPackageError("module did not pass sanity validation", reason="Invalid semantic version")

    if StrictVersion(metadata["min_redis_pack_version"]) < StrictVersion(module_metadata.MIN_REDIS_PACK_VERSION):
        raise UnpackerPackageError("module did not pass sanity validation", reason="Min redis pack version is too low")

    # wrong signature
    # TODO: this check should be deffered to a later stage
    # As _sha256_checksum will read entire module file
    # And we're unable to seek it back to its starting point.
    # if _sha256_checksum(module) != metadata["sha256"]:
    #     raise UnpackerPackageError("module did not pass sanity validation", reason="Wrong signature")

    return True



import os
import hashlib
import platform

# Defaults
ARCHITECTURE = platform.machine()
OS = platform.system()
DISPLAY_NAME = ""
AUTHOR = ""
EMAIL = ""
DESCRIPTION = ""
HOMEPAGE = ""
LICENSE = ""
MODULE_NAME = ""
MODULE_VERSION = 1
MODULE_SEMANTIC_VERSION = '0.0.1'
MODULE_COMMANDS = []
EXCLUDE_COMMANDS = []
MODULE_CAPABILITIES = []
COMMAND_LINE_ARGS = ""
MIN_REDIS_VERSION = "4.0"
MIN_REDIS_PACK_VERSION = "5.0"
RAMP_FORMAT_VERSION = 1
CONFIG_COMMAND = ""
OVERIDE_COMMAND = []

FIELDS = [
    "architecture",
    "author",
    "capabilities",
    "command_line_args",
    "commands",
    "config_command",
    "description",
    "display_name",
    "email",
    "exclude_commands",
    "git_sha",
    "homepage",
    "license",
    "min_redis_pack_version",
    "min_redis_version",
    "module_file",
    "module_name",
    "os",
    "overide_command",
    "ramp_format_version",
    "semantic_version",
    "sha256",
    "version",
]

def sha256_checksum(filename, block_size=65536):
    """Computes sha256 for given file"""
    sha256 = hashlib.sha256()
    with open(filename, 'rb') as f:
        for block in iter(lambda: f.read(block_size), b''):
            sha256.update(block)
    return sha256.hexdigest()

def create_default_metadata(module_path):
    """Creates a default metadata"""
    metadata = {
        "module_name" : MODULE_NAME,
        "module_file": os.path.basename(module_path),
        "os": OS,
        "architecture": ARCHITECTURE,
        "version": MODULE_VERSION,
        "semantic_version": MODULE_SEMANTIC_VERSION,
        "display_name": DISPLAY_NAME,
        "author": AUTHOR,
        "email": EMAIL,
        "description": DESCRIPTION,
        "homepage": HOMEPAGE,
        "license": LICENSE,
        "command_line_args": COMMAND_LINE_ARGS,
        "capabilities": MODULE_CAPABILITIES,
        "min_redis_version": MIN_REDIS_VERSION,
        "min_redis_pack_version": MIN_REDIS_PACK_VERSION,
        "sha256": sha256_checksum(module_path),
        "commands": MODULE_COMMANDS,
        "ramp_format_version": RAMP_FORMAT_VERSION,
        "config_command": CONFIG_COMMAND,
        "exclude_commands": EXCLUDE_COMMANDS,
        "overide_command": OVERIDE_COMMAND
    }
    return metadata



from setuptools import setup, find_packages
from RAMP.version import VERSION

LONG_DESC = """
# RAMP
Redis Automatic Module Packaging

Similar to `npm init`, this packer bundles Redis Modules for later distribution.

It gathers information from modules e.g.
module's name, command list, version and additional metadata.

# Prerequisites
Make sure redis-server is on your PATH

```sh
export PATH=$PATH:<PATH_TO_REDIS>
```

# Usage

## Command line mode

```sh
ramp <PATH_TO_RedisModule.so> -a <author> -e <email> -ar <architecture> -d <description> -ho <homepage> -l <license> -ex <extras> -c <cmdargs> -r <redis-min-version>
```

For Help

```sh
ramp -h
```

## Output
ramp generates module.zip

Which contains:

    1. RedisModule.so - original module
    2. Module.json - module's metadata

"""

setup(
    name='ramp-packer',
    py_modules=['RAMP'],
    version=str(VERSION),
    description='Packs for Redis modules into a distributable format',
    author='RedisLabs',
    url='https://github.com/redislabs/RAMP',
    download_url='https://github.com/RedisLabs/RAMP/archive/{}.tar.gz'.format(VERSION),
    license='BSD 2-clause',
    long_description=LONG_DESC,
    packages=find_packages(),
    install_requires=['redis<=2.10.6', 'pyyaml', 'click>=6.7', 'semantic_version==2.6.0'],
    entry_points='''
        [console_scripts]
        ramp=RAMP.ramp:ramp
    ''',
)
